import enum


class CommonStereotype(enum.Enum):
    IMPLEMENTATION = 1
    BIND = 2
    DERIVE = 3
    FRIEND = 4
    INSTANCEOF = 5
    INSTANTIATE = 6
    REFINE = 7
    USE = 8
    SUBSTITUTE = 9
    ACCESS = 10
    IMPORT = 11
    PERMIT = 12
    EXTEND = 13
    INCLUDE = 14
    BECOME = 15
    CALL = 16
    COPY = 17
    PARAMETER = 18
    SEND = 19


Stereotype = str



"""
Contains the code generation interface
"""
import abc
from typing import List, Tuple

from plantuml_creator import settings
from plantuml_creator.diagram_type import DiagramType

from plantuml_creator.uml_code import PlantUMLCode, CodeStyle
from plantuml_creator.error import Error


class CodeGenerator(metaclass=abc.ABCMeta):
    """
    It is the generation interface.
    The classes that perform it must generate valid plantuml code
    """
    
    class Context(object):
        """
        The generation context
        """
        
        def __init__(self, parent: 'CodeGenerator' = None, indent_level: int = 0,
                     diagram_type: DiagramType = DiagramType.UNKNOWN,
                     code_style: CodeStyle = settings.CODE_STYLE):
            self.code_style = code_style
            self.diagram_type = diagram_type
            self.parent = parent
            self.indent_level = indent_level
        
        def increase_indent_level(self) -> int:
            self.indent_level += 1
            return self.indent_level
    
    """
    Class that must generate valid plantuml code
    """
    
    @abc.abstractmethod
    def gen_code(self, ctx: Context) -> Tuple[PlantUMLCode, List[Error]]:
        """
        This method when implemented must generate valid plantuml code.
        :return: A tuple where:
        El primer elemento será un código de plantuml válido
        The second element will be a list of errors that occurred during the code generation process
        """
        raise NotImplementedError("")



import enum
from typing import Union, Tuple, Optional

import stringcase


class Color(enum.Enum):
    ALICE_BLUE = "#F0F8FF"
    ANTIQUE_WHITE = "#FAEBD7"
    AQUA = "#00FFFF"
    AQUAMARINE = "#7FFFD4"
    AZURE = "#F0FFFF"
    BEIGE = "#F5F5DC"
    BISQUE = "#FFE4C4"
    BLACK = "#000000"
    BLANCHED_ALMOND = "#FFEBCD"
    BLUE = "#0000FF"
    BLUE_VIOLET = "#8A2BE2"
    BROWN = "#A52A2A"
    BURLY_WOOD = "#DEB887"
    CADET_BLUE = "#5F9EA0"
    CHARTREUSE = "#7FFF00"
    CHOCOLATE = "#D2691E"
    CORAL = "#FF7F50"
    CORNFLOWER_BLUE = "#6495ED"
    CORNSILK = "#FFF8DC"
    CRIMSON = "#DC143C"
    CYAN = "#00FFFF"
    DARK_BLUE = "#00008B"
    DARK_CYAN = "#008B8B"
    DARK_GOLDEN_ROD = "#B8860B"
    DARK_GRAY = "#A9A9A9"
    DARK_GREY = "#A9A9A9"
    DARK_GREEN = "#006400"
    DARK_KHAKI = "#BDB76B"
    DARK_MAGENTA = "#8B008B"
    DARK_OLIVE_GREEN = "#556B2F"
    DARKORANGE = "#FF8C00"
    DARK_ORCHID = "#9932CC"
    DARK_RED = "#8B0000"
    DARK_SALMON = "#E9967A"
    DARK_SEA_GREEN = "#8FBC8F"
    DARK_SLATE_BLUE = "#483D8B"
    DARK_SLATE_GRAY = "#2F4F4F"
    DARK_SLATE_GREY = "#2F4F4F"
    DARK_TURQUOISE = "#00CED1"
    DARK_VIOLET = "#9400D3"
    DEEP_PINK = "#FF1493"
    DEEP_SKY_BLUE = "#00BFFF"
    DIM_GRAY = "#696969"
    DIM_GREY = "#696969"
    DODGER_BLUE = "#1E90FF"
    FIRE_BRICK = "#B22222"
    FLORAL_WHITE = "#FFFAF0"
    FOREST_GREEN = "#228B22"
    FUCHSIA = "#FF00FF"
    GAINSBORO = "#DCDCDC"
    GHOST_WHITE = "#F8F8FF"
    GOLD = "#FFD700"
    GOLDEN_ROD = "#DAA520"
    GRAY = "#808080"
    GREY = "#808080"
    GREEN = "#008000"
    GREEN_YELLOW = "#ADFF2F"
    HONEY_DEW = "#F0FFF0"
    HOT_PINK = "#FF69B4"
    INDIAN_RED = "#CD5C5C"
    INDIGO = "#4B0082"
    IVORY = "#FFFFF0"
    KHAKI = "#F0E68C"
    LAVENDER = "#E6E6FA"
    LAVENDER_BLUSH = "#FFF0F5"
    LAWN_GREEN = "#7CFC00"
    LEMON_CHIFFON = "#FFFACD"
    LIGHT_BLUE = "#ADD8E6"
    LIGHT_CORAL = "#F08080"
    LIGHT_CYAN = "#E0FFFF"
    LIGHT_GOLDEN_ROD_YELLOW = "#FAFAD2"
    LIGHT_GRAY = "#D3D3D3"
    LIGHT_GREY = "#D3D3D3"
    LIGHT_GREEN = "#90EE90"
    LIGHT_PINK = "#FFB6C1"
    LIGHT_SALMON = "#FFA07A"
    LIGHT_SEA_GREEN = "#20B2AA"
    LIGHT_SKY_BLUE = "#87CEFA"
    LIGHT_SLATE_GRAY = "#778899"
    LIGHT_SLATE_GREY = "#778899"
    LIGHT_STEEL_BLUE = "#B0C4DE"
    LIGHT_YELLOW = "#FFFFE0"
    LIME = "#00FF00"
    LIME_GREEN = "#32CD32"
    LINEN = "#FAF0E6"
    MAGENTA = "#FF00FF"
    MAROON = "#800000"
    MEDIUM_AQUA_MARINE = "#66CDAA"
    MEDIUM_BLUE = "#0000CD"
    MEDIUM_ORCHID = "#BA55D3"
    MEDIUM_PURPLE = "#9370D8"
    MEDIUM_SEA_GREEN = "#3CB371"
    MEDIUM_SLATE_BLUE = "#7B68EE"
    MEDIUM_SPRING_GREEN = "#00FA9A"
    MEDIUM_TURQUOISE = "#48D1CC"
    MEDIUM_VIOLET_RED = "#C71585"
    MIDNIGHT_BLUE = "#191970"
    MINT_CREAM = "#F5FFFA"
    MISTY_ROSE = "#FFE4E1"
    MOCCASIN = "#FFE4B5"
    NAVAJO_WHITE = "#FFDEAD"
    NAVY = "#000080"
    OLD_LACE = "#FDF5E6"
    OLIVE = "#808000"
    OLIVE_DRAB = "#6B8E23"
    ORANGE = "#FFA500"
    ORANGE_RED = "#FF4500"
    ORCHID = "#DA70D6"
    PALE_GOLDEN_ROD = "#EEE8AA"
    PALE_GREEN = "#98FB98"
    PALE_TURQUOISE = "#AFEEEE"
    PALE_VIOLET_RED = "#D87093"
    PAPAYA_WHIP = "#FFEFD5"
    PEACH_PUFF = "#FFDAB9"
    PERU = "#CD853F"
    PINK = "#FFC0CB"
    PLUM = "#DDA0DD"
    POWDER_BLUE = "#B0E0E6"
    PURPLE = "#800080"
    RED = "#FF0000"
    ROSY_BROWN = "#BC8F8F"
    ROYAL_BLUE = "#4169E1"
    SADDLE_BROWN = "#8B4513"
    SALMON = "#FA8072"
    SANDY_BROWN = "#F4A460"
    SEA_GREEN = "#2E8B57"
    SEA_SHELL = "#FFF5EE"
    SIENNA = "#A0522D"
    SILVER = "#C0C0C0"
    SKY_BLUE = "#87CEEB"
    SLATE_BLUE = "#6A5ACD"
    SLATE_GRAY = "#708090"
    SLATE_GREY = "#708090"
    SNOW = "#FFFAFA"
    SPRING_GREEN = "#00FF7F"
    STEEL_BLUE = "#4682B4"
    TAN = "#D2B48C"
    TEAL = "#008080"
    THISTLE = "#D8BFD8"
    TOMATO = "#FF6347"
    TURQUOISE = "#40E0D0"
    VIOLET = "#EE82EE"
    WHEAT = "#F5DEB3"
    WHITE = "#FFFFFF"
    WHITE_SMOKE = "#F5F5F5"
    YELLOW = "#FFFF00"
    YELLOW_GREEN = "#9ACD32"
    # Archimate
    BUSINESS = "#FFFF00"
    APPLICATION = "#A9DCDF"
    MOTIVATION = "#B19CD9"
    STRATEGY = "#F6E4CC"
    TECHNOLOGY = "#90EE90"
    PHYSICAL = "#CCFFCC"
    IMPLEMENTATION = "#FFA6BF"


class ColorHelper(object):
    @staticmethod
    def normalize(color: Union[Color, str]) -> Tuple[str, Optional[Exception]]:
        """
        Return a normalized hexadecimal color
        
        :param color: valid values:
                      - Valids colors enums values in:
                            * lower or upper snake case. i.e: "blue" "dark_salmon"
                            * upper camel case. i.e: "DarkSalmon" "Blue"
                      - Hexadecimal strings with "#" at begin or not
                            * #F0F8FF
                            * F0F8FF
                       
        :return: A color in hexadecimal format with # in front. i.e: "#F0F8FF"
        """
        
        if color is None or len(color) == 0:
            return "", None
        if isinstance(color, Color):
            return color.value, None
        
        # i.e: "blue", "Aqua" "dark_salmon" or "DarkSalmon"
        color_skc = stringcase.snakecase(color).upper()
        if color_skc in Color.__dict__:
            return Color.__dict__[color_skc]
        
        # i.e: '#F0F8FF'
        if color.startswith("#"):
            return color, None
        
        # i.e: 'F0F8FF'
        try:
            _ = int(color[1:], 16)
            return f"#{color}", None
        except ValueError:
            pass
        return "", ValueError(f"invalid color: {color}")



import enum
from typing import Tuple, List, Dict

from plantuml_creator.uml_code import PlantUMLCode
from plantuml_creator.code_generator import CodeGenerator
from plantuml_creator.error import Error
from plantuml_creator.stereotype import Stereotype
from plantuml_creator.uml_object import ObjectName, UMLObj, UMLObjType

Quantifier = Tuple[str, str]


class RelationType(enum.Enum):
    Association = 1
    ReflexiveAssociation = 2
    
    Dependency = 4
    
    Realization = 6
    RealizationCanonical = 7
    RealizationElided = 8
    
    Composition = 9
    Aggregation = 10
    Extension = 11
    Call = 12


class Orientation(enum.Enum):
    UNDEFINED = ""
    UP = "up"
    DOWN = "down"
    LEFT = "left"
    RIGHT = "right"


class Relation(UMLObj):
    default_shape = "-{}-"
    default_stereotype_start = "<<"
    default_stereotype_end = ">>"
    relation_shape_map: Dict[RelationType, str] = {
        # Principals
        RelationType.Aggregation: "o-{}-",
        RelationType.Composition: "*-{}-",
        RelationType.Extension: "-{}-|>",
        RelationType.Association: "-{}-",
        
        RelationType.ReflexiveAssociation: "-{}-",
        
        RelationType.Dependency: "-{}->",
        RelationType.Call: "-{}->",
        
        RelationType.Realization: ".{}..|>",
        RelationType.RealizationCanonical: ".{}.|>",
        RelationType.RealizationElided: ".{}.|>",  # TODO: Change to circle
    }
    
    def __init__(self, obj1: ObjectName, rel_type: RelationType, obj2: ObjectName,
                 stereotypes: List[Stereotype] = None, quantifiers: Quantifier = None,
                 orientation: Orientation = Orientation.UNDEFINED):
        super().__init__("", UMLObjType.RELATION, {})
        self.orientation = orientation
        if quantifiers is None:
            quantifiers = ("", "")
        if stereotypes is None:
            stereotypes = []
        self.quantifiers = quantifiers
        self.stereotypes = stereotypes
        self.obj2 = obj2
        self.rel_type = rel_type
        self.obj1 = obj1
    
    def gen_code(self, ctx: CodeGenerator.Context) -> Tuple[PlantUMLCode, List[Error]]:
        """
        Generate the code of a relationship
        :return: i.e:
        Relation(
            obj1="person",
            rel_type=RelationType.Association,
            obj2="account",
            stereotypes=["use"],
            quantifiers=("1", "*")
        ).gen_code()
        >> person "1" -- "*" account : <<use>>
        """
        shape = Relation.default_shape
        if self.rel_type in Relation.relation_shape_map:
            shape = Relation.relation_shape_map[self.rel_type]
        if len(self.stereotypes) > 0:
            stereotypes = f": {self._format_stereotypes(self.stereotypes)}"
        else:
            stereotypes = ""
        q1, q2 = self._format_quantifier(self.quantifiers)
        
        shape = shape.replace("{}", str(self.orientation.value))
        out = f"{self.obj1} {q1} {shape} {q2} {self.obj2}{stereotypes}"
        # Remove all duplicates spaces
        out = " ".join(out.split()).strip()
        return out, []
    
    @staticmethod
    def _format_quantifier(quantifier: Quantifier) -> Quantifier:
        q1 = f"\"{quantifier[0]}\"" if quantifier[0] is not None and len(quantifier[0]) > 0 else ""
        q2 = f"\"{quantifier[1]}\"" if quantifier[1] is not None and len(quantifier[1]) > 0 else ""
        return q1, q2
    
    @staticmethod
    def _format_stereotypes(stereotypes: List[Stereotype]) -> str:
        out_st = ""
        for i in range(len(stereotypes)):
            st = stereotypes[i]
            out_st += f"{Relation.default_stereotype_start}{st}{Relation.default_stereotype_end}"
            if i != len(stereotypes) - 1:
                out_st += "\n"
        return out_st



import enum

DataType = str


class CommonDataType(enum.Enum):
    STRING = DataType("String")
    INT = DataType("int")
    FLOAT = DataType("float")
    DOUBLE = DataType("double")
    BOOLEAN = DataType("boolean")
    MAP = DataType("Map")
    SET = DataType("Set")
    List = DataType("List")



import enum


class DiagramType(enum.Enum):
    UNKNOWN = 0
    SEQUENCE = 1
    CLASS = 2
    USE_CASE = 3
    ACTIVITY = 4
    COMPONENT = 5
    STATE = 5
    OBJECT = 6
    DEPLOY = 7
    TIMING = 8
    
    # Non UML
    
    WIREFRAME = 9
    ARCHIMATE = 10
    SDL = 11
    DITAA = 12
    GANTT = 13
    MINDMAP = 14
    WORK_BREAKDOWN_STRUCTURE = 15
    MATH = 16
    ENTITY_RELATIONSHIP = 17


import enum
from typing import List, Tuple, Dict, Union

from plantuml_creator.code_generator import CodeGenerator
from plantuml_creator.diagram_type import DiagramType
from plantuml_creator.error import Error
from plantuml_creator.shape import ShapeAttributes
from plantuml_creator.uml_code import PlantUMLCode
from plantuml_creator.uml_object import UMLObj, UMLObjType
from plantuml_creator.utils import ListJoiner, JoinItem

OBJECTS_SEPARATOR = "/' OBJECTS '/"
RELATION_SEPARATOR = "/' RELATIONS '/"

DEFAULT_ADD_SEPARATORS = True


class DiagramOptName(enum.Enum):
    AUTO_NUMBER = 1
    AUTO_NUMBER_START = 2
    TITLE = 5
    # Not supported
    HEADER_PAGE = 3
    FOOTER_PAGE = 4
    # If is true then adds the OBJECTS_SEPARATOR and RELATION_SEPARATOR
    # To the plantuml code
    ADD_SEPARATORS = 6


DiagramOptValue = Union[str, bool, int, float]

DiagramOpts = Dict[DiagramOptName, DiagramOptValue]

Wrapper = Tuple[str, str]


class Diagram(UMLObj):
    """
    Return a valid UML diagram
    """
    
    standard_wrapper = ("@startuml", "@enduml")
    wireframe_wrapper = ("@startsalt", "@endsalt")
    work_breakdown_wrapper = ("@startwbs", "@endwbs")
    mindmap_wrapper = ("@startmindmap", "@endmindmap")
    gantt_wrapper = ("@startgantt", "@endgantt")
    ditaa_wrapper = ("@startditaa", "@endditaa")
    
    standard_headers_by_types = [
        DiagramType.SEQUENCE,
        DiagramType.USE_CASE,
        DiagramType.CLASS,
        DiagramType.ACTIVITY,
        DiagramType.COMPONENT,
        DiagramType.STATE,
        DiagramType.OBJECT,
        DiagramType.DEPLOY,
        DiagramType.TIMING,
        DiagramType.ENTITY_RELATIONSHIP,
        DiagramType.MATH,
        DiagramType.ARCHIMATE,
        DiagramType.TIMING,
        DiagramType.DEPLOY,
    
    ]
    
    def __init__(self, name: str, opts: DiagramOpts, type_: DiagramType, objs: List[UMLObj], relations: List[UMLObj],
                 attr: ShapeAttributes = None):
        super().__init__(name, UMLObjType.DIAGRAM, attr)
        if opts is None:
            opts = {}
        self.opts = opts
        self.objs = objs
        if relations is None:
            relations = []
        if objs is None:
            objs = []
        self.objects = objs
        self.relations = relations
        self._wrapper: Wrapper = self._get_wrapper(type_)
        self.type = type_
        if DiagramOptName.ADD_SEPARATORS not in self.opts and DEFAULT_ADD_SEPARATORS:
            self.opts[DiagramOptName.ADD_SEPARATORS] = True
    
    @classmethod
    def _get_wrapper(cls, type_: DiagramType) -> Wrapper:
        if type_ in cls.standard_headers_by_types:
            return cls.standard_wrapper
        if type_ in DiagramType.WIREFRAME:
            return cls.wireframe_wrapper
        if type_ in DiagramType.WORK_BREAKDOWN_STRUCTURE:
            return cls.work_breakdown_wrapper
        if type_ in DiagramType.MINDMAP:
            return cls.mindmap_wrapper
        if type_ in DiagramType.GANTT:
            return cls.gantt_wrapper
        if type_ in DiagramType.DITAA:
            return cls.ditaa_wrapper
        return cls.standard_wrapper
    
    @classmethod
    def _generate_optionals(cls, type_: DiagramType, opts: DiagramOpts) -> PlantUMLCode:
        out = ""
        if DiagramOptName.TITLE in opts:
            out += f'title "{opts[DiagramOptName.TITLE]}"\n'
        auto_number_start = ""
        if type_ == DiagramType.SEQUENCE:
            if DiagramOptName.AUTO_NUMBER_START in opts:
                auto_number_start = opts[DiagramOptName.AUTO_NUMBER_START]
            if opts[DiagramOptName.AUTO_NUMBER] or len(auto_number_start) > 0:
                out += f'autonumber {auto_number_start}\n'
        return out
    
    def _generate_header(self) -> PlantUMLCode:
        optionals = self._generate_optionals(self.type, self.opts)
        return f"{self._wrapper[0]}\n" \
               f"{optionals}\n"
    
    def gen_code(self, ctx: CodeGenerator.Context) -> Tuple[PlantUMLCode, List[Error]]:
        """
        Generate the code for the entire diagram
        """
        ctx.increase_indent_level()
        d, err = ListJoiner.join({
            "objs": JoinItem(self.objs, "\n", None, "", ctx),
            "relations": JoinItem(self.relations, "\n", None, "", ctx),
        })
        header = self._generate_header()
        obj_sep = ""
        rel_sep = ""
        if DiagramOptName.ADD_SEPARATORS in self.opts and self.opts[DiagramOptName.ADD_SEPARATORS]:
            obj_sep = OBJECTS_SEPARATOR
            rel_sep = RELATION_SEPARATOR
        return f'{header}\n' \
               f'{obj_sep}\n' \
               f'{d["objs"]}\n' \
               f'{rel_sep}\n' \
               f'{d["relations"]}\n' \
               f'{self._wrapper[1]}\n', err



import re
from typing import List, Tuple

from plantuml_creator.code_generator import CodeGenerator
from plantuml_creator.uml_code import PlantUMLCode
from plantuml_creator.error import Error
from plantuml_creator.relation import Relation
from plantuml_creator.shape import ShapeAttributes, ContainerShapeType
from plantuml_creator.uml_object import UMLObj, UMLObjType
from plantuml_creator.utils import ListJoiner, JoinItem

ContainerContent = Tuple[List[UMLObj], List[Relation]]


class Container(UMLObj):
    """
    Model a uml class object.
    It generates a code like the following:
    class ClassName {
        /' Attributes '/
        String a
        String b

        /' Methods '/
        void methodA()
        void methodB()
    }
    """
    
    def __init__(self, name: str,
                 objs: List[UMLObj],
                 container_type: ContainerShapeType = ContainerShapeType.PACKAGE,
                 shape_attrs: ShapeAttributes = None):
        super().__init__(name, UMLObjType.CONTAINER, shape_attrs)
        self.objs = objs
        self.container_type = container_type
    
    def gen_code(self, ctx: CodeGenerator.Context) -> Tuple[PlantUMLCode, List[Error]]:
        d, err = ListJoiner.join({
            "objs": JoinItem(self.objs, "\n    ", None, "", ctx),
        })
        syb_type = ""
        if self.container_type in [ContainerShapeType.NAMESPACE,
                                   ContainerShapeType.CLOUD,
                                   ContainerShapeType.DATABASE]:
            type_ = self.container_type.value
        else:
            type_ = "package"
            if self.container_type != ContainerShapeType.PACKAGE:
                syb_type = f"<<{self.container_type.value}>>"
        
        color, err_color = self.extract_color_by_shape_attrs(self.shape_attrs)
        header = re.sub(' +', ' ', f"{type_} \"{self.name}\" {syb_type} {color} {{").strip()
        out = f"{header}\n" \
              f"    {d['objs']}\n" \
              f"}}"
        return out, err



import abc
import enum

from typing import List, Tuple, Optional

from plantuml_creator.color import ColorHelper
from plantuml_creator.uml_code import PlantUMLCode
from plantuml_creator.code_generator import CodeGenerator
from plantuml_creator.error import Error
from plantuml_creator.shape import ShapeAttributes, ShapeAttributeName

ObjectName = str


class UMLObjType(enum.Enum):
    CLASS = 1
    USE_CASE = 2
    CONTAINER = 3
    COMPONENT = 4
    DIAGRAM = 5
    VARIABLE = 6
    METHOD = 7
    RELATION = 8


class UMLObj(CodeGenerator):
    allowed_variable_chars = ["A"]
    name_replacement_char = "_"
    
    def __init__(self, name: str, obj_type: UMLObjType, shape_attrs: ShapeAttributes):
        super().__init__()
        if name is None or len(name) == 0:
            name = " "
        if shape_attrs is None:
            shape_attrs = {}
        self.shape_attrs = shape_attrs
        self.obj_type = obj_type
        self.name = name
    
    @staticmethod
    def is_a_to_z(c: str) -> bool:
        return (ord("a") <= ord(c) <= ord("z")) or (ord("A") <= ord(c) <= ord("Z"))
    
    @classmethod
    def convert_to_variable_name(cls, name: str) -> str:
        out = ""
        for c in name:
            if c.isnumeric() or cls.is_a_to_z(c):
                out += c
            else:
                out += cls.name_replacement_char
        return out
    
    @staticmethod
    def extract_color_by_shape_attrs(shape_attrs: ShapeAttributes) -> Tuple[str, Optional[Exception]]:
        if ShapeAttributeName.COLOR not in shape_attrs:
            return "", None
        color, err_color = ColorHelper.normalize(shape_attrs[ShapeAttributeName.COLOR])
        if err_color is not None:
            return "", err_color
        return color, None
    
    @staticmethod
    def gen_for_generators_list(generators: List[List[CodeGenerator]]):
        out_codes: PlantUMLCode = ""
        out_errors = []
        for gen in generators:
            for gen_i in gen:
                code_lines, obj_errors = gen_i.gen_code()
                out_codes = code_lines + "\n"
                if len(obj_errors) > 0:
                    out_errors.extend(out_errors)
        return out_codes, out_errors
    
    @abc.abstractmethod
    def gen_code(self, ctx: CodeGenerator.Context) -> Tuple[PlantUMLCode, List[Error]]:
        """
        Generate the code for all objects and relationships that are in the diagram.
        If any object or relationship has an error, then add it
        """
        raise NotImplementedError()



import abc
import enum
from typing import Callable, Dict

from plantuml_creator.data_types import DataType

COMMENT_START = "/'"
COMMENT_END = "/'"
OBJ_CODE_HEADER = f"\n{COMMENT_START} ---- OBJECTS --- {COMMENT_END}\n"
REL_CODE_HEADER = f"\n{COMMENT_START} ---- RELATIONS --- {COMMENT_END}\n"

PlantUMLCode = str


class CodeStyle(enum.Enum):
    C = 1  # C/C++, Java, Javascript, Typescript, etc
    PYTHON = 2
    GOLANG = 3


class _CodeStyleChangerBase(metaclass=abc.ABCMeta):
    
    @staticmethod
    @abc.abstractmethod
    def type_normalization(data_type: str) -> str:
        raise NotImplementedError()


class _PythonCodeStyleChanger(_CodeStyleChangerBase):
    primitive_names_map = {
        "string": "str",
        "integer": "int",
        "double": "float",
        "boolean": "bool",
    }
    
    @staticmethod
    def type_normalization(data_type: str) -> str:
        if data_type.lower() in _PythonCodeStyleChanger.primitive_names_map:
            return _PythonCodeStyleChanger.primitive_names_map[data_type.lower()]
        return data_type


class _GolangCodeStyleChanger(_CodeStyleChangerBase):
    primitive_names_map = {
        "String": "string",
        "Integer": "int",
        "Double": "float",
        "Boolean": "bool",
        "boolean": "bool",
    }
    
    @staticmethod
    def type_normalization(data_type: str) -> str:
        if data_type in _GolangCodeStyleChanger.primitive_names_map:
            return _GolangCodeStyleChanger.primitive_names_map[data_type]
        return data_type


class CodeStyleChanger(metaclass=abc.ABCMeta):
    styles_map: Dict[CodeStyle, Callable[[str], str]] = {
        CodeStyle.PYTHON: _PythonCodeStyleChanger.type_normalization,
        CodeStyle.GOLANG: _GolangCodeStyleChanger.type_normalization,
    }
    
    @classmethod
    def type_normalization(cls, data_type: DataType, new_style: CodeStyle) -> str:
        """
        Normalize the name of the data type according to the language style.
        In case the style is not supported, then returns the original string
        """
        if new_style in cls.styles_map:
            return cls.styles_map[new_style](data_type)
        return data_type



import enum
from typing import Union

StorageClassSpecifier = str
TypeQualifier = str
AccessModifier = str


class CommonStorageClassSpecifier(enum.Enum):
    """
    A storage class specifier is used to refine the declaration of a variable, a function,
    and parameters. Storage classes determine whether:
    The object has internal, external, or no linkage
    The object is to be stored in memory or in a register, if available
    The object receives the default initial value of 0 or an indeterminate default initial value
    The object can be referenced throughout a program or only within the function, block, or source
        file where the variable is defined
    The storage duration for the object is maintained throughout program run time or only during
        the execution of the block where the object is defined
    """
    AUTO = StorageClassSpecifier("auto")
    STATIC = StorageClassSpecifier("static")
    EXTERN = StorageClassSpecifier("extern")
    MUTABLE = StorageClassSpecifier("mutable")
    REGISTER = StorageClassSpecifier("register")


class CommonTypeQualifier(enum.Enum):
    CONST = TypeQualifier("const")
    UNSIGNED = TypeQualifier("unsigned")
    FINAL = TypeQualifier("final")
    VOLATILE = TypeQualifier("volatile")
    RESTRICT = TypeQualifier("restrict")
    ATOMIC = TypeQualifier("atomic")
    IMMUTABLE = TypeQualifier("immutable")
    SHARED = TypeQualifier("shared")
    INOUT = TypeQualifier("inout")
    SYNC = TypeQualifier("sync")
    ASYNC = TypeQualifier("async")


class CommonAccessModifier(enum.Enum):
    EMPTY = AccessModifier("")
    DEFAULT = AccessModifier("default")
    PRIVATE = AccessModifier("private")
    PROTECTED = AccessModifier("protected")
    PUBLIC = AccessModifier("public")
    FRIEND = AccessModifier("friend")

    
TypeModifier = Union[TypeQualifier, StorageClassSpecifier, AccessModifier]



from typing import Tuple, List, Callable, Dict, Union

import stringcase

from plantuml_creator.uml_code import PlantUMLCode, CodeStyle, CodeStyleChanger
from plantuml_creator.code_generator import CodeGenerator
from plantuml_creator.data_types import DataType
from plantuml_creator.error import Error
from plantuml_creator.modifier import TypeModifier
from plantuml_creator.uml_object import UMLObj, UMLObjType


class Variable(UMLObj):
    """
    Specify a variable declaration.
    It can mean a parameter or variable in any context
    
    example 1:
    in the following function:
    void f( const int a)
    The variable would be the "a" parameter
    
    example 2:
    in the following declaration:
    static const a;
    The variable would be the "a" parameter
    """
    
    _generators = {}
    
    def __init__(self, name: str, type_: DataType,  modifiers: List[TypeModifier] = None):
        super().__init__(name, UMLObjType.VARIABLE, {})
        if modifiers is None:
            modifiers = []
        self.type_ = type_
        self.modifiers = modifiers
        
    
    @classmethod
    def _get_generators(cls) -> Dict[CodeStyle, Callable[['Variable'], Tuple[PlantUMLCode, List[Error]]]]:
        if len(cls._generators.keys()) == 0:
            cls._generators = {
                CodeStyle.C: cls.gen_code_c_style,
                CodeStyle.GOLANG: cls.gen_code_golang_style,
                CodeStyle.PYTHON: cls.gen_code_python_style,
            }
        return cls._generators
    
    @staticmethod
    def gen_code_c_style(var: 'Variable') -> Tuple[PlantUMLCode, List[Error]]:
        """
        Generate c-style code:
        :return:
        i.e:
        "static int a";
        or
        "const String a";
        """
        modifiers = ""
        for mod in var.modifiers:
            if mod is not None:
                modifiers += mod + " "
        type_ = CodeStyleChanger.type_normalization(var.type_, CodeStyle.C)
        return f"{modifiers}{type_} {var.name}", []
    
    @staticmethod
    def gen_code_python_style(var: 'Variable') -> Tuple[PlantUMLCode, List[Error]]:
        """
        Generate c-style code:
        :return:
        i.e:
        "a: int";
        or
        "a: str";
        """
        # Note: the modifiers are omitted when you have the python style
        name = stringcase.snakecase(var.name)
        type_ = CodeStyleChanger.type_normalization(var.type_, CodeStyle.PYTHON)
        return f"{name}: {type_}", []
    
    @staticmethod
    def gen_code_golang_style(var: 'Variable') -> Tuple[PlantUMLCode, List[Error]]:
        """
        Generate c-style code:
        :return:
        i.e:
        "a int";
        or
        "a string";
        """
        # Note: the modifiers are omitted when you have the golang style
        type_ = CodeStyleChanger.type_normalization(var.type_, CodeStyle.GOLANG)
        return f"{var.name} {type_}", []
    
    def gen_code(self, ctx: CodeGenerator.Context) -> Tuple[PlantUMLCode, List[Error]]:
        return Variable._get_generators()[ctx.code_style](self)



from typing import Tuple, List, Dict, Callable

from plantuml_creator.data_types import DataType
from plantuml_creator.modifier import TypeModifier, AccessModifier, CommonAccessModifier
from plantuml_creator.uml_code import PlantUMLCode, CodeStyle
from plantuml_creator.code_generator import CodeGenerator
from plantuml_creator.error import Error
from plantuml_creator.uml_object import UMLObj, UMLObjType
from plantuml_creator.utils import ListJoiner, JoinItem
from plantuml_creator.variable import Variable

_DEFAULT_ARGS_BOUNDS = ("(", ")")


class Method(UMLObj):
    """
    Model a method or function
    i.e:
                     args
    return name     ___|___
     |      |      |       |
     v      v      v       v
    String foo(String a, int b)
    """
    _generators = {}
    
    def __init__(self, name: str, return_types: List[DataType], args: List[Variable],
                 modifiers: List[TypeModifier] = None):
        super().__init__(name, UMLObjType.METHOD, {})
        if modifiers is None:
            modifiers = []
        self.modifiers = modifiers
        self.args = args
        self.return_types = return_types
    
    @classmethod
    def _get_generators(cls) -> Dict[CodeStyle, Callable[['Method'], Tuple[PlantUMLCode, List[Error]]]]:
        if len(cls._generators.keys()) == 0:
            cls._generators = {
                CodeStyle.C: cls.gen_code_c_style,
                CodeStyle.GOLANG: cls.gen_code_golang_style,
                CodeStyle.PYTHON: cls.gen_code_python_style,
            }
        return cls._generators
    
    @staticmethod
    def gen_code_all_style(method: 'Method',
                           ctx: CodeGenerator.Context, ret_sep=None) -> Tuple[Dict[str, str], List[Error]]:
        if ret_sep is None:
            ret_sep = ("", "")
        return ListJoiner.join({
            "modifiers": JoinItem(method.modifiers, " ", ("", ""), "", ctx),
            "returns": JoinItem(method.return_types, ",", ret_sep, "", ctx),
            "args": JoinItem(method.args, ", ", _DEFAULT_ARGS_BOUNDS, "()", ctx),
        })
    
    @classmethod
    def gen_code_c_style(cls, method: 'Method', ctx: CodeGenerator.Context) -> Tuple[PlantUMLCode, List[Error]]:
        """
        Generate c-style code:
        :return:
        i.e:
        "public static String foo(String a, int b)";
        """
        d, errs = cls.gen_code_all_style(method, ctx=ctx)
        return f"{d['modifiers']} {d['returns']} {method.name}{d['args']}", errs
    
    @classmethod
    def gen_code_python_style(cls, method: 'Method', ctx: CodeGenerator.Context) -> Tuple[PlantUMLCode, List[Error]]:
        """
        Generate python-style code:
        :return:
        i.e:
        "def foo(a: str, b: int) -> String";
        """
        d, errs = cls.gen_code_all_style(method, ctx)
        if d['returns'] is None or len(d['returns']) == "":
            return_body = ""
        else:
            if d['returns'] == "void":
                d['returns'] = "None"
            return_body = f" -> {d['returns']}"
        if len(method.return_types) == 1:
            return_body = return_body.replace("(", "").replace(")", "")
        return f"def {method.name}{d['args']}{return_body}:", errs
    
    @classmethod
    def gen_code_golang_style(cls, method: 'Method', ctx: CodeGenerator.Context) -> Tuple[PlantUMLCode, List[Error]]:
        """
        Generate go-style code:
        :return:
        i.e:
        "func foo(a string, b int) string";
        """
        # Note: the modifiers are omitted when you have the golang style
        d, errs = cls.gen_code_all_style(method, ctx, ret_sep=("(", ")"))
        if len(method.return_types) == 1:
            d['returns'] = d['returns'].replace("(", "").replace(")", "")
        method.name = method.name[0].upper() + method.name[1:]
        if method.modifiers is not None:
            for i in range(len(method.modifiers)):
                if isinstance(method.modifiers[i], AccessModifier) and \
                        method.modifiers[i] in [CommonAccessModifier.PRIVATE.value, CommonAccessModifier.PROTECTED]:
                    method.name = method.name[0].lower() + method.name[1:]
        return f"func {method.name}{d['args']} {d['returns']}", errs
    
    def gen_code(self, ctx: CodeGenerator.Context) -> Tuple[PlantUMLCode, List[Error]]:
        return Method._get_generators()[ctx.code_style](self, ctx)



from typing import List, Tuple

from plantuml_creator.code_generator import CodeGenerator
from plantuml_creator.uml_code import PlantUMLCode
from plantuml_creator.error import Error
from plantuml_creator.shape import ShapeAttributes, ShapeAttributeName, ContainerShapeType
from plantuml_creator.uml_object import UMLObj, UMLObjType


class Component(UMLObj):
    """
    i.e:
    component "<b>BS_UT_BCO_PropertyHandler.biz</b>" as BS_UT_BCO_PropertyHandler.biz #LightGray
    """
    
    def __init__(self, name: str,
                 shape_attrs: ShapeAttributes = None):
        super().__init__(name, UMLObjType.COMPONENT, shape_attrs)
    
    @staticmethod
    def get_component_diagram(attr_name: str) -> str:
        if attr_name in [ContainerShapeType.DATABASE.value]:
            return ContainerShapeType.DATABASE.value
        return attr_name
    
    def gen_code(self, ctx: CodeGenerator.Context) -> Tuple[PlantUMLCode, List[Error]]:
        name_to_show = self.name.replace("\"", self.name_replacement_char)
        color, err = self.extract_color_by_shape_attrs(self.shape_attrs)
        shape_type = "component"
        if ShapeAttributeName.CONTAINER_SHAPE_TYPE in self.shape_attrs:
            shape_type = self.shape_attrs[ShapeAttributeName.CONTAINER_SHAPE_TYPE]
            shape_type = self.get_component_diagram(shape_type)
        if err is not None:
            color = ""
        
        return f'{shape_type} "{name_to_show}" as {self.convert_to_variable_name(self.name)} {color}', []



"""
Setup
"""
import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="plantuml_creator",
    version="1.0.2",
    description="Library to generate plantuml code from python objects",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/Eitol/plantuml_creator",
    author="Hector Oliveros",
    author_email="hector.oliveros.leon@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
    ],
    packages=["plantuml_creator"],
    include_package_data=True,
    install_requires=[
        "attrs==19.3.0",
        "importlib-metadata==1.3.0",
        "more-itertools==8.0.2",
        "packaging==20.0",
        "pluggy==0.13.1",
        "py==1.8.1",
        "pyparsing==2.4.6",
        "pytest==5.3.2",
        "six==1.13.0",
        "str2bool==1.1",
        "stringcase==1.2.0",
        "wcwidth==0.1.8",
        "zipp==0.6.0",
    ],
    entry_points={
        "console_scripts": []
    },
)



import enum
from typing import List, Tuple

from plantuml_creator.code_generator import CodeGenerator
from plantuml_creator.color import ColorHelper

from plantuml_creator.uml_code import PlantUMLCode
from plantuml_creator.error import Error
from plantuml_creator.method import Method
from plantuml_creator.shape import ShapeAttributes, ShapeAttributeName
from plantuml_creator.uml_object import UMLObj, UMLObjType
from plantuml_creator.utils import ListJoiner, JoinItem
from plantuml_creator.variable import Variable


class ClassType(enum.Enum):
    REGULAR_CLASS = "class"
    ABSTRACT = "abstract class"
    INTERFACE = "interface"
    INTERFACE_LOLLIPOP = "interface"
    
    def __str__(self):
        return ""


class Class(UMLObj):
    ADD_DIVIDER = True
    ATTRIBUTES_ZONE_DIVIDER = r"/' Attributes '/"
    METHODS_ZONE_DIVIDER = r"/' Methods '/"
    
    """
    Model a uml class object.
    It generates a code like the following:
    class ClassName {
        /' Attributes '/
        String a
        String b
        
        /' Methods '/
        void methodA()
        void methodB()
    }
    """
    
    def __init__(self, name: str, type_: ClassType = ClassType.REGULAR_CLASS,
                 attrs: List[Variable] = None,
                 methods: List[Method] = None,
                 shape_attrs: ShapeAttributes = None):
        super().__init__(name, UMLObjType.CLASS, shape_attrs)
        self.type_ = type_
        if attrs is None:
            attrs = []
        if methods is None:
            methods = []
        self.attrs = attrs
        self.methods = methods
    
    def gen_lollipop(self, color: str = "") -> str:
        """
        interface "CctFccActTipEnt" as CctFccActTipEnt #red
        :return:
        """
        alias = self.convert_to_variable_name(self.name)
        if color != "":
            color = " " + color
        return f'interface "{self.name}" as {alias}{color}'
    
    def gen_code(self, ctx: CodeGenerator.Context) -> Tuple[PlantUMLCode, List[Error]]:
        err: List[Error] = []
        style = ""
        if ShapeAttributeName.COLOR in self.shape_attrs:
            color, err_color = ColorHelper.normalize(self.shape_attrs[ShapeAttributeName.COLOR])
            if err_color is not None:
                err.extend(err)
            else:
                style += color
        if self.type_ == ClassType.INTERFACE_LOLLIPOP:
            return self.gen_lollipop(style), []
        
        d, err = ListJoiner.join({
            "attrs": JoinItem(self.attrs, "\n    ", None, "", ctx),
            "methods": JoinItem(self.methods, "\n    ", None, "", ctx),
        })
        
        return f"{str(self.type_.value)} {self.name} {style} {{\n" \
               f"    {self.ATTRIBUTES_ZONE_DIVIDER}\n" \
               f"    {d['attrs']}\n\n" \
               f"    {self.METHODS_ZONE_DIVIDER}\n" \
               f"    {d['methods']}\n" \
               f"}}\n", err



import enum
from typing import Dict

ShapeAttributeValue = str


class ContainerShapeType(enum.Enum):
    NAMESPACE = "namespace"
    NODE = "node"
    PACKAGE = "package"
    CLOUD = "cloud"
    DATABASE = "database"
    FRAME = "frame"
    FOLDER = "folder"


class ShapeAttributeName(enum.Enum):
    COLOR = 1
    CONTAINER_SHAPE_TYPE = 2


ShapeAttributes = Dict[ShapeAttributeName, ShapeAttributeValue]



from typing import Any, Tuple, Union, List, Dict, Optional

from plantuml_creator.code_generator import CodeGenerator
from plantuml_creator.data_types import DataType
from plantuml_creator.error import Error
from plantuml_creator.uml_code import CodeStyleChanger

EMPTY_BOUND = ("", "")


class JoinItem(object):
    def __init__(self, list_: Any, sep: str, bounds: Optional[Tuple[str, str]],
                 default_: str, ctx: CodeGenerator.Context):
        self.ctx = ctx
        self.default_ = default_
        if bounds is None:
            bounds = EMPTY_BOUND
        self.bounds = bounds
        self.sep = sep
        self.list_ = list_


class ListJoiner(object):
    
    @staticmethod
    def to_str(obj: Any, ctx: CodeGenerator.Context) -> Tuple[str, List[Error]]:
        if isinstance(obj, CodeGenerator):
            return obj.gen_code(ctx)
        if isinstance(obj, DataType):
            return CodeStyleChanger.type_normalization(obj, ctx.code_style), []
        return str(obj), []
    
    @staticmethod
    def _join(it: JoinItem) -> Tuple[str, List[Error]]:
        if it.list_ is None or len(it.list_) == 0:
            return (f"{it.bounds[0]}{it.bounds[1]}" if it.default_ is None else it.default_), []
        out, errors = ListJoiner.to_str(it.list_[0], it.ctx)
        if len(it.list_) > 1:
            for i in range(1, len(it.list_)):
                out_aux, errors_aux = ListJoiner.to_str(it.list_[i], it.ctx)
                if len(errors_aux) > 0:
                    errors.extend(errors_aux)
                if len(out_aux) > 0:
                    out += it.sep + out_aux
        return f"{it.bounds[0]}{out}{it.bounds[1]}", errors
    
    @classmethod
    def join(cls, items: Dict[str, JoinItem]) -> Tuple[Dict[str, str], List[Error]]:
        out_val: Dict[str, str] = {}
        out_errs = []
        for name, it in items.items():
            str_val, errs = cls._join(it)
            out_errs.extend(errs)
            out_val[name] = str_val
        return out_val, out_errs
